@@ -32,7 +32,7 @@ const struct cmd_entry cmd_find_window_entry = {
 	.name = "find-window",
 	.alias = "findw",
 
-	.args = { "CiNrt:TZ", 1, 1 },
+	.args = { "CiNrt:TZ", 1, 1, NULL },
 	.usage = "[-CiNrTZ] " CMD_TARGET_PANE_USAGE " match-string",
 
 	.target = { 't', CMD_FIND_PANE, 0 },
@@ -47,8 +47,8 @@ cmd_find_window_exec(struct cmd *self, struct cmdq_item *item)
 	struct args		*args = cmd_get_args(self), *new_args;
 	struct cmd_find_state	*target = cmdq_get_target(item);
 	struct window_pane	*wp = target->wp;
-	const char		*s = args->argv[0], *suffix = "";
-	char			*filter, *argv = { NULL };
+	const char		*s = args_string(args, 0), *suffix = "";
+	struct args_value	*filter;
 	int			 C, N, T;
 
 	C = args_has(args, 'C');
@@ -65,41 +65,49 @@ cmd_find_window_exec(struct cmd *self, struct cmdq_item *item)
 	if (!C && !N && !T)
 		C = N = T = 1;
 
+	filter = xcalloc(1, sizeof *filter);
+	filter->type = ARGS_STRING;
+
 	if (C && N && T) {
-		xasprintf(&filter,
+		xasprintf(&filter->string,
 		    "#{||:"
 		    "#{C%s:%s},#{||:#{m%s:*%s*,#{window_name}},"
 		    "#{m%s:*%s*,#{pane_title}}}}",
 		    suffix, s, suffix, s, suffix, s);
 	} else if (C && N) {
-		xasprintf(&filter,
+		xasprintf(&filter->string,
 		    "#{||:#{C%s:%s},#{m%s:*%s*,#{window_name}}}",
 		    suffix, s, suffix, s);
 	} else if (C && T) {
-		xasprintf(&filter,
+		xasprintf(&filter->string,
 		    "#{||:#{C%s:%s},#{m%s:*%s*,#{pane_title}}}",
 		    suffix, s, suffix, s);
 	} else if (N && T) {
-		xasprintf(&filter,
+		xasprintf(&filter->string,
 		    "#{||:#{m%s:*%s*,#{window_name}},"
 		    "#{m%s:*%s*,#{pane_title}}}",
 		    suffix, s, suffix, s);
-	} else if (C)
-		xasprintf(&filter, "#{C%s:%s}", suffix, s);
-	else if (N)
-		xasprintf(&filter, "#{m%s:*%s*,#{window_name}}", suffix, s);
-	else
-		xasprintf(&filter, "#{m%s:*%s*,#{pane_title}}", suffix, s);
+	} else if (C) {
+		xasprintf(&filter->string,
+		    "#{C%s:%s}",
+		    suffix, s);
+	} else if (N) {
+		xasprintf(&filter->string,
+		    "#{m%s:*%s*,#{window_name}}",
+		    suffix, s);
+	} else {
+		xasprintf(&filter->string,
+		    "#{m%s:*%s*,#{pane_title}}",
+		    suffix, s);
+	}
 
-	new_args = args_parse("", 1, &argv);
+	new_args = args_create();
 	if (args_has(args, 'Z'))
 		args_set(new_args, 'Z', NULL);
 	args_set(new_args, 'f', filter);
 
 	window_pane_set_mode(wp, NULL, &window_tree_mode, target, new_args);
-
 	args_free(new_args);
-	free(filter);
 
 	return (CMD_RETURN_NORMAL);
 }
