@@ -30,7 +30,10 @@
  * Runs a command without a window.
  */
 
-static enum cmd_retval	cmd_run_shell_exec(struct cmd *, struct cmdq_item *);
+static enum args_parse_type	cmd_run_shell_args_parse(struct args *, u_int,
+				    char **);
+static enum cmd_retval		cmd_run_shell_exec(struct cmd *,
+				    struct cmdq_item *);
 
 static void	cmd_run_shell_timer(int, short, void *);
 static void	cmd_run_shell_callback(struct job *);
@@ -41,7 +44,7 @@ const struct cmd_entry cmd_run_shell_entry = {
 	.name = "run-shell",
 	.alias = "run",
 
-	.args = { "bd:Ct:", 0, 1 },
+	.args = { "bd:Ct:", 0, 1, cmd_run_shell_args_parse },
 	.usage = "[-bC] [-d delay] " CMD_TARGET_PANE_USAGE " [shell-command]",
 
 	.target = { 't', CMD_FIND_PANE, CMD_FIND_CANFAIL },
@@ -51,18 +54,26 @@ const struct cmd_entry cmd_run_shell_entry = {
 };
 
 struct cmd_run_shell_data {
-	struct client		*client;
-	char			*cmd;
-	int			 shell;
-	char			*cwd;
-	struct cmdq_item	*item;
-	struct session		*s;
-	int			 wp_id;
-	struct event		 timer;
-	int			 flags;
-	struct cmd_parse_input	 pi;
+	struct client			*client;
+	char				*cmd;
+	struct args_command_state	*state;
+	char				*cwd;
+	struct cmdq_item		*item;
+	struct session			*s;
+	int				 wp_id;
+	struct event			 timer;
+	int				 flags;
 };
 
+static enum args_parse_type
+cmd_run_shell_args_parse(struct args *args, __unused u_int idx,
+    __unused char **cause)
+{
+	if (args_has(args, 'C'))
+		return (ARGS_PARSE_COMMANDS_OR_STRING);
+	return (ARGS_PARSE_STRING);
+}
+
 static void
 cmd_run_shell_print(struct job *job, const char *msg)
 {
@@ -73,22 +84,17 @@ cmd_run_shell_print(struct job *job, const char *msg)
 
 	if (cdata->wp_id != -1)
 		wp = window_pane_find_by_id(cdata->wp_id);
-	if (wp == NULL) {
-		if (cdata->item != NULL) {
-			cmdq_print(cdata->item, "%s", msg);
-			return;
-		}
-		if (cmd_find_from_nothing(&fs, 0) != 0)
-			return;
+	if (wp == NULL && cdata->item != NULL)
+		wp = server_client_get_pane(cdata->client);
+	if (wp == NULL && cmd_find_from_nothing(&fs, 0) == 0)
 		wp = fs.wp;
-		if (wp == NULL)
-			return;
-	}
+	if (wp == NULL)
+		return;
 
 	wme = TAILQ_FIRST(&wp->modes);
 	if (wme == NULL || wme->mode != &window_view_mode)
 		window_pane_set_mode(wp, NULL, &window_view_mode, NULL, NULL);
-	window_copy_add(wp, "%s", msg);
+	window_copy_add(wp, 1, "%s", msg);
 }
 
 static enum cmd_retval
@@ -100,7 +106,7 @@ cmd_run_shell_exec(struct cmd *self, struct cmdq_item *item)
 	struct client			*tc = cmdq_get_target_client(item);
 	struct session			*s = target->s;
 	struct window_pane		*wp = target->wp;
-	const char			*delay;
+	const char			*delay, *cmd;
 	double				 d;
 	struct timeval			 tv;
 	char				*end;
@@ -112,21 +118,17 @@ cmd_run_shell_exec(struct cmd *self, struct cmdq_item *item)
 			cmdq_error(item, "invalid delay time: %s", delay);
 			return (CMD_RETURN_ERROR);
 		}
-	} else if (args->argc == 0)
+	} else if (args_count(args) == 0)
 		return (CMD_RETURN_NORMAL);
 
 	cdata = xcalloc(1, sizeof *cdata);
-	if (args->argc != 0)
-		cdata->cmd = format_single_from_target(item, args->argv[0]);
-
-	cdata->shell = !args_has(args, 'C');
-	if (!cdata->shell) {
-		memset(&cdata->pi, 0, sizeof cdata->pi);
-		cmd_get_source(self, &cdata->pi.file, &cdata->pi.line);
-		if (wait)
-			cdata->pi.item = item;
-		cdata->pi.c = tc;
-		cmd_find_copy_state(&cdata->pi.fs, target);
+	if (!args_has(args, 'C')) {
+		cmd = args_string(args, 0);
+		if (cmd != NULL)
+			cdata->cmd = format_single_from_target(item, cmd);
+	} else {
+		cdata->state = args_make_commands_prepare(self, item, 0, NULL,
+		    wait, 1);
 	}
 
 	if (args_has(args, 't') && wp != NULL)
@@ -170,34 +172,38 @@ cmd_run_shell_timer(__unused int fd, __unused short events, void* arg)
 	struct cmd_run_shell_data	*cdata = arg;
 	struct client			*c = cdata->client;
 	const char			*cmd = cdata->cmd;
+	struct cmdq_item		*item = cdata->item, *new_item;
+	struct cmd_list			*cmdlist;
 	char				*error;
-	struct cmdq_item		*item = cdata->item;
-	enum cmd_parse_status		 status;
 
-	if (cmd != NULL && cdata->shell) {
-		if (job_run(cmd, 0, NULL, cdata->s, cdata->cwd, NULL,
+	if (cdata->state == NULL) {
+		if (cmd == NULL) {
+			if (cdata->item != NULL)
+				cmdq_continue(cdata->item);
+			cmd_run_shell_free(cdata);
+			return;
+		}
+		if (job_run(cmd, 0, NULL, NULL, cdata->s, cdata->cwd, NULL,
 		    cmd_run_shell_callback, cmd_run_shell_free, cdata,
 		    cdata->flags, -1, -1) == NULL)
 			cmd_run_shell_free(cdata);
 		return;
 	}
 
-	if (cmd != NULL) {
-		if (item != NULL) {
-			status = cmd_parse_and_insert(cmd, &cdata->pi, item,
-			    cmdq_get_state(item), &error);
-		} else {
-			status = cmd_parse_and_append(cmd, &cdata->pi, c, NULL,
-			    &error);
-		}
-		if (status == CMD_PARSE_ERROR) {
-			if (cdata->item == NULL) {
-				*error = toupper((u_char)*error);
-				status_message_set(c, -1, 1, 0, "%s", error);
-			} else
-				cmdq_error(cdata->item, "%s", error);
-			free(error);
-		}
+	cmdlist = args_make_commands(cdata->state, 0, NULL, &error);
+	if (cmdlist == NULL) {
+		if (cdata->item == NULL) {
+			*error = toupper((u_char)*error);
+			status_message_set(c, -1, 1, 0, "%s", error);
+		} else
+			cmdq_error(cdata->item, "%s", error);
+		free(error);
+	} else if (item == NULL) {
+		new_item = cmdq_get_command(cmdlist, NULL);
+		cmdq_append(c, new_item);
+	} else {
+		new_item = cmdq_get_command(cmdlist, cmdq_get_state(item));
+		cmdq_insert_after(item, new_item);
 	}
 
 	if (cdata->item != NULL)
@@ -216,7 +222,8 @@ cmd_run_shell_callback(struct job *job)
 	int				 retcode, status;
 
 	do {
-		if ((line = evbuffer_readline(event->input)) != NULL) {
+		line = evbuffer_readln(event->input, NULL, EVBUFFER_EOL_LF);
+		if (line != NULL) {
 			cmd_run_shell_print(job, line);
 			free(line);
 		}
@@ -265,6 +272,8 @@ cmd_run_shell_free(void *data)
 		session_remove_ref(cdata->s, __func__);
 	if (cdata->client != NULL)
 		server_client_unref(cdata->client);
+	if (cdata->state != NULL)
+		args_make_commands_free(cdata->state);
 	free(cdata->cwd);
 	free(cdata->cmd);
 	free(cdata);
