@@ -80,8 +80,11 @@ screen_init(struct screen *s, u_int sx, u_int sy, u_int hlimit)
 	s->titles = NULL;
 	s->path = NULL;
 
-	s->cstyle = 0;
-	s->ccolour = xstrdup("");
+	s->cstyle = SCREEN_CURSOR_DEFAULT;
+	s->default_cstyle = SCREEN_CURSOR_DEFAULT;
+	s->default_mode = 0;
+	s->ccolour = -1;
+	s->default_ccolour = -1;
 	s->tabs = NULL;
 	s->sel = NULL;
 
@@ -100,7 +103,7 @@ screen_reinit(struct screen *s)
 	s->rupper = 0;
 	s->rlower = screen_size_y(s) - 1;
 
-	s->mode = MODE_CURSOR|MODE_WRAP;
+	s->mode = MODE_CURSOR|MODE_WRAP|(s->mode & MODE_CRLF);
 	if (options_get_number(global_options, "extended-keys") == 2)
 		s->mode |= MODE_KEXTENDED;
 
@@ -125,7 +128,6 @@ screen_free(struct screen *s)
 	free(s->tabs);
 	free(s->path);
 	free(s->title);
-	free(s->ccolour);
 
 	if (s->write_list != NULL)
 		screen_write_free_list(s);
@@ -151,22 +153,47 @@ screen_reset_tabs(struct screen *s)
 		bit_set(s->tabs, i);
 }
 
-/* Set screen cursor style. */
+/* Set screen cursor style and mode. */
 void
-screen_set_cursor_style(struct screen *s, u_int style)
+screen_set_cursor_style(u_int style, enum screen_cursor_style *cstyle,
+    int *mode)
 {
-	if (style <= 6) {
-		s->cstyle = style;
-		s->mode &= ~MODE_BLINKING;
+	switch (style) {
+	case 0:
+		*cstyle = SCREEN_CURSOR_DEFAULT;
+		break;
+	case 1:
+		*cstyle = SCREEN_CURSOR_BLOCK;
+		*mode |= MODE_CURSOR_BLINKING;
+		break;
+	case 2:
+		*cstyle = SCREEN_CURSOR_BLOCK;
+		*mode &= ~MODE_CURSOR_BLINKING;
+		break;
+	case 3:
+		*cstyle = SCREEN_CURSOR_UNDERLINE;
+		*mode |= MODE_CURSOR_BLINKING;
+		break;
+	case 4:
+		*cstyle = SCREEN_CURSOR_UNDERLINE;
+		*mode &= ~MODE_CURSOR_BLINKING;
+		break;
+	case 5:
+		*cstyle = SCREEN_CURSOR_BAR;
+		*mode |= MODE_CURSOR_BLINKING;
+		break;
+	case 6:
+		*cstyle = SCREEN_CURSOR_BAR;
+		*mode &= ~MODE_CURSOR_BLINKING;
+		break;
 	}
 }
 
 /* Set screen cursor colour. */
 void
-screen_set_cursor_colour(struct screen *s, const char *colour)
+screen_set_cursor_colour(struct screen *s, int colour)
 {
-	free(s->ccolour);
-	s->ccolour = xstrdup(colour);
+	s->ccolour = colour;
 }
 
 /* Set screen title. */
@@ -626,3 +653,53 @@ screen_alternate_off(struct screen *s, struct grid_cell *gc, int cursor)
 	if (s->cy > screen_size_y(s) - 1)
 		s->cy = screen_size_y(s) - 1;
 }
+
+/* Get mode as a string. */
+const char *
+screen_mode_to_string(int mode)
+{
+	static char	tmp[1024];
+
+	if (mode == 0)
+		return ("NONE");
+	if (mode == ALL_MODES)
+		return ("ALL");
+
+	*tmp = '\0';
+	if (mode & MODE_CURSOR)
+		strlcat(tmp, "CURSOR,", sizeof tmp);
+	if (mode & MODE_INSERT)
+		strlcat(tmp, "INSERT,", sizeof tmp);
+	if (mode & MODE_KCURSOR)
+		strlcat(tmp, "KCURSOR,", sizeof tmp);
+	if (mode & MODE_KKEYPAD)
+		strlcat(tmp, "KKEYPAD,", sizeof tmp);
+	if (mode & MODE_WRAP)
+		strlcat(tmp, "WRAP,", sizeof tmp);
+	if (mode & MODE_MOUSE_STANDARD)
+		strlcat(tmp, "MOUSE_STANDARD,", sizeof tmp);
+	if (mode & MODE_MOUSE_BUTTON)
+		strlcat(tmp, "MOUSE_BUTTON,", sizeof tmp);
+	if (mode & MODE_CURSOR_BLINKING)
+		strlcat(tmp, "CURSOR_BLINKING,", sizeof tmp);
+	if (mode & MODE_CURSOR_VERY_VISIBLE)
+		strlcat(tmp, "CURSOR_VERY_VISIBLE,", sizeof tmp);
+	if (mode & MODE_MOUSE_UTF8)
+		strlcat(tmp, "UTF8,", sizeof tmp);
+	if (mode & MODE_MOUSE_SGR)
+		strlcat(tmp, "SGR,", sizeof tmp);
+	if (mode & MODE_BRACKETPASTE)
+		strlcat(tmp, "BRACKETPASTE,", sizeof tmp);
+	if (mode & MODE_FOCUSON)
+		strlcat(tmp, "FOCUSON,", sizeof tmp);
+	if (mode & MODE_MOUSE_ALL)
+		strlcat(tmp, "MOUSE_ALL,", sizeof tmp);
+	if (mode & MODE_ORIGIN)
+		strlcat(tmp, "ORIGIN,", sizeof tmp);
+	if (mode & MODE_CRLF)
+		strlcat(tmp, "CRLF,", sizeof tmp);
+	if (mode & MODE_KEXTENDED)
+		strlcat(tmp, "KEXTENDED,", sizeof tmp);
+	tmp[strlen(tmp) - 1] = '\0';
+	return (tmp);
+}
