@@ -18,13 +18,14 @@
 
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/utsname.h>
 
 #include <errno.h>
-#include <event.h>
 #include <fcntl.h>
 #include <langinfo.h>
 #include <locale.h>
 #include <pwd.h>
+#include <signal.h>
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
@@ -45,15 +46,15 @@ const char	*shell_command;
 static __dead void	 usage(void);
 static char		*make_label(const char *, char **);
 
+static int		 areshell(const char *);
 static const char	*getshell(void);
-static int		 checkshell(const char *);
 
 static __dead void
 usage(void)
 {
 	fprintf(stderr,
-	    "usage: %s [-2CluvV] [-c shell-command] [-f file] [-L socket-name]\n"
-	    "            [-S socket-path] [command [flags]]\n",
+	    "usage: %s [-2CDlNuvV] [-c shell-command] [-f file] [-L socket-name]\n"
+	    "            [-S socket-path] [-T features] [command [flags]]\n",
 	    getprogname());
 	exit(1);
 }
@@ -75,7 +76,7 @@ getshell(void)
 	return (_PATH_BSHELL);
 }
 
-static int
+int
 checkshell(const char *shell)
 {
 	if (shell == NULL || *shell != '/')
@@ -87,7 +88,7 @@ checkshell(const char *shell)
 	return (1);
 }
 
-int
+static int
 areshell(const char *shell)
 {
 	const char	*progname, *ptr;
@@ -104,47 +105,133 @@ areshell(const char *shell)
 	return (0);
 }
 
+static char *
+expand_path(const char *path, const char *home)
+{
+	char			*expanded, *name;
+	const char		*end;
+	struct environ_entry	*value;
+
+	if (strncmp(path, "~/", 2) == 0) {
+		if (home == NULL)
+			return (NULL);
+		xasprintf(&expanded, "%s%s", home, path + 1);
+		return (expanded);
+	}
+
+	if (*path == '$') {
+		end = strchr(path, '/');
+		if (end == NULL)
+			name = xstrdup(path + 1);
+		else
+			name = xstrndup(path + 1, end - path - 1);
+		value = environ_find(global_environ, name);
+		free(name);
+		if (value == NULL)
+			return (NULL);
+		if (end == NULL)
+			end = "";
+		xasprintf(&expanded, "%s%s", value->value, end);
+		return (expanded);
+	}
+
+	return (xstrdup(path));
+}
+
+static void
+expand_paths(const char *s, char ***paths, u_int *n, int ignore_errors)
+{
+	const char	*home = find_home();
+	char		*copy, *next, *tmp, resolved[PATH_MAX], *expanded;
+	char		*path;
+	u_int		 i;
+
+	*paths = NULL;
+	*n = 0;
+
+	copy = tmp = xstrdup(s);
+	while ((next = strsep(&tmp, ":")) != NULL) {
+		expanded = expand_path(next, home);
+		if (expanded == NULL) {
+			log_debug("%s: invalid path: %s", __func__, next);
+			continue;
+		}
+		if (realpath(expanded, resolved) == NULL) {
+			log_debug("%s: realpath(\"%s\") failed: %s", __func__,
+			    expanded, strerror(errno));
+			if (ignore_errors) {
+				free(expanded);
+				continue;
+			}
+			path = expanded;
+		} else {
+			path = xstrdup(resolved);
+			free(expanded);
+		}
+		for (i = 0; i < *n; i++) {
+			if (strcmp(path, (*paths)[i]) == 0)
+				break;
+		}
+		if (i != *n) {
+			log_debug("%s: duplicate path: %s", __func__, path);
+			free(path);
+			continue;
+		}
+		*paths = xreallocarray(*paths, (*n) + 1, sizeof *paths);
+		(*paths)[(*n)++] = path;
+	}
+	free(copy);
+}
+
 static char *
 make_label(const char *label, char **cause)
 {
-	char		*base, resolved[PATH_MAX], *path, *s;
-	struct stat	 sb;
-	uid_t		 uid;
+	char		**paths, *path, *base;
+	u_int		  i, n;
+	struct stat	  sb;
+	uid_t		  uid;
 
 	*cause = NULL;
-
 	if (label == NULL)
 		label = "default";
 	uid = getuid();
 
-	if ((s = getenv("TMUX_TMPDIR")) != NULL && *s != '\0')
-		xasprintf(&base, "%s/tmux-%ld", s, (long)uid);
-	else
-		xasprintf(&base, "%s/tmux-%ld", _PATH_TMP, (long)uid);
-	if (realpath(base, resolved) == NULL &&
-	    strlcpy(resolved, base, sizeof resolved) >= sizeof resolved) {
-		errno = ERANGE;
-		free(base);
+	expand_paths(TMUX_SOCK, &paths, &n, 1);
+	if (n == 0) {
+		xasprintf(cause, "no suitable socket path");
+		return (NULL);
+	}
+	path = paths[0]; /* can only have one socket! */
+	for (i = 1; i < n; i++)
+		free(paths[i]);
+	free(paths);
+
+	xasprintf(&base, "%s/tmux-%ld", path, (long)uid);
+	free(path);
+	if (mkdir(base, S_IRWXU) != 0 && errno != EEXIST) {
+		xasprintf(cause, "couldn't create directory %s (%s)", base,
+		    strerror(errno));
 		goto fail;
 	}
-
-	if (mkdir(resolved, S_IRWXU) != 0 && errno != EEXIST)
-		goto fail;
-	if (lstat(resolved, &sb) != 0)
+	if (lstat(base, &sb) != 0) {
+		xasprintf(cause, "couldn't read directory %s (%s)", base,
+		    strerror(errno));
 		goto fail;
+	}
 	if (!S_ISDIR(sb.st_mode)) {
-		errno = ENOTDIR;
+		xasprintf(cause, "%s is not a directory", base);
 		goto fail;
 	}
 	if (sb.st_uid != uid || (sb.st_mode & S_IRWXO) != 0) {
-		errno = EACCES;
+		xasprintf(cause, "directory %s has unsafe permissions", base);
 		goto fail;
 	}
-	xasprintf(&path, "%s/%s", resolved, label);
+	xasprintf(&path, "%s/%s", base, label);
+	free(base);
 	return (path);
 
 fail:
-	xasprintf(cause, "error creating %s (%s)", resolved, strerror(errno));
+	free(base);
 	return (NULL);
 }
 
@@ -162,6 +249,33 @@ setblocking(int fd, int state)
 	}
 }
 
+uint64_t
+get_timer(void)
+{
+	struct timespec	ts;
+
+	/*
+	 * We want a timestamp in milliseconds suitable for time measurement,
+	 * so prefer the monotonic clock.
+	 */
+	if (clock_gettime(CLOCK_MONOTONIC, &ts) != 0)
+		clock_gettime(CLOCK_REALTIME, &ts);
+	return ((ts.tv_sec * 1000ULL) + (ts.tv_nsec / 1000000ULL));
+}
+
+const char *
+sig2name(int signo)
+{
+     static char	s[11];
+
+#ifdef HAVE_SYS_SIGNAME
+     if (signo > 0 && signo < NSIG)
+	     return (sys_signame[signo]);
+#endif
+     xsnprintf(s, sizeof s, "%d", signo);
+     return (s);
+}
+
 const char *
 find_cwd(void)
 {
@@ -208,13 +322,22 @@ find_home(void)
 	return (home);
 }
 
+const char *
+getversion(void)
+{
+	return (TMUX_VERSION);
+}
+
 int
 main(int argc, char **argv)
 {
-	char					*path, *label, *cause, **var;
-	const char				*s, *shell, *cwd;
-	int					 opt, flags, keys;
+	char					*path = NULL, *label = NULL;
+	char					*cause, **var;
+	const char				*s, *cwd;
+	int					 opt, keys, feat = 0, fflag = 0;
+	uint64_t				 flags = 0;
 	const struct options_table_entry	*oe;
+	u_int					 i;
 
 	if (setlocale(LC_CTYPE, "en_US.UTF-8") == NULL &&
 	    setlocale(LC_CTYPE, "C.UTF-8") == NULL) {
@@ -230,30 +353,46 @@ main(int argc, char **argv)
 
 	if (**argv == '-')
 		flags = CLIENT_LOGIN;
-	else
-		flags = 0;
 
-	label = path = NULL;
-	while ((opt = getopt(argc, argv, "2c:Cdf:lL:qS:uUVv")) != -1) {
+	global_environ = environ_create();
+	for (var = environ; *var != NULL; var++)
+		environ_put(global_environ, *var, 0);
+	if ((cwd = find_cwd()) != NULL)
+		environ_set(global_environ, "PWD", 0, "%s", cwd);
+	expand_paths(TMUX_CONF, &cfg_files, &cfg_nfiles, 1);
+
+	while ((opt = getopt(argc, argv, "2c:CDdf:lL:NqS:T:uUvV")) != -1) {
 		switch (opt) {
 		case '2':
-			flags |= CLIENT_256COLOURS;
+			tty_add_features(&feat, "256", ":,");
 			break;
 		case 'c':
 			shell_command = optarg;
 			break;
+		case 'D':
+			flags |= CLIENT_NOFORK;
+			break;
 		case 'C':
 			if (flags & CLIENT_CONTROL)
 				flags |= CLIENT_CONTROLCONTROL;
 			else
 				flags |= CLIENT_CONTROL;
 			break;
-		case 'V':
-			printf("%s %s\n", getprogname(), VERSION);
-			exit(0);
 		case 'f':
-			set_cfg_file(optarg);
+			if (!fflag) {
+				fflag = 1;
+				for (i = 0; i < cfg_nfiles; i++)
+					free(cfg_files[i]);
+				cfg_nfiles = 0;
+			}
+			cfg_files = xreallocarray(cfg_files, cfg_nfiles + 1,
+			    sizeof *cfg_files);
+			cfg_files[cfg_nfiles++] = xstrdup(optarg);
+			cfg_quiet = 0;
 			break;
+ 		case 'V':
+			printf("%s %s\n", getprogname(), getversion());
+ 			exit(0);
 		case 'l':
 			flags |= CLIENT_LOGIN;
 			break;
@@ -261,12 +400,18 @@ main(int argc, char **argv)
 			free(label);
 			label = xstrdup(optarg);
 			break;
+		case 'N':
+			flags |= CLIENT_NOSTARTSERVER;
+			break;
 		case 'q':
 			break;
 		case 'S':
 			free(path);
 			path = xstrdup(optarg);
 			break;
+		case 'T':
+			tty_add_features(&feat, optarg, ":,");
+			break;
 		case 'u':
 			flags |= CLIENT_UTF8;
 			break;
@@ -282,6 +427,8 @@ main(int argc, char **argv)
 
 	if (shell_command != NULL && argc != 0)
 		usage();
+	if ((flags & CLIENT_NOFORK) && argc != 0)
+		usage();
 
 	if ((ptm_fd = getptmfd()) == -1)
 		err(1, "getptmfd");
@@ -311,12 +458,6 @@ main(int argc, char **argv)
 			flags |= CLIENT_UTF8;
 	}
 
-	global_environ = environ_create();
-	for (var = environ; *var != NULL; var++)
-		environ_put(global_environ, *var);
-	if ((cwd = find_cwd()) != NULL)
-		environ_set(global_environ, "PWD", "%s", cwd);
-
 	global_options = options_create(NULL);
 	global_s_options = options_create(NULL);
 	global_w_options = options_create(NULL);
@@ -333,11 +474,12 @@ main(int argc, char **argv)
 	 * The default shell comes from SHELL or from the user's passwd entry
 	 * if available.
 	 */
-	shell = getshell();
-	options_set_string(global_s_options, "default-shell", 0, "%s", shell);
+	options_set_string(global_s_options, "default-shell", 0, "%s",
+	    getshell());
 
 	/* Override keys to vi if VISUAL or EDITOR are set. */
 	if ((s = getenv("VISUAL")) != NULL || (s = getenv("EDITOR")) != NULL) {
+		options_set_string(global_options, "editor", 0, "%s", s);
 		if (strrchr(s, '/') != NULL)
 			s = strrchr(s, '/') + 1;
 		if (strstr(s, "vi") != NULL)
@@ -360,16 +502,19 @@ main(int argc, char **argv)
 			path[strcspn(path, ",")] = '\0';
 		}
 	}
-	if (path == NULL && (path = make_label(label, &cause)) == NULL) {
-		if (cause != NULL) {
-			fprintf(stderr, "%s\n", cause);
-			free(cause);
+	if (path == NULL) {
+		if ((path = make_label(label, &cause)) == NULL) {
+			if (cause != NULL) {
+				fprintf(stderr, "%s\n", cause);
+				free(cause);
+			}
+			exit(1);
 		}
-		exit(1);
+		flags |= CLIENT_DEFAULTSOCKET;
 	}
 	socket_path = path;
 	free(label);
 
 	/* Pass control to the client. */
-	exit(client_main(osdep_event_init(), argc, argv, flags));
+	exit(client_main(osdep_event_init(), argc, argv, flags, feat));
 }
