@@ -308,7 +308,7 @@ window_buffer_get_key(void *modedata, void *itemdata, u_int line)
 	}
 	pb = paste_get_name(item->name);
 	if (pb == NULL)
-		return KEYC_NONE;
+		return (KEYC_NONE);
 
 	ft = format_create(NULL, NULL, FORMAT_NONE, 0);
 	format_defaults(ft, NULL, NULL, 0, NULL);
@@ -320,7 +320,7 @@ window_buffer_get_key(void *modedata, void *itemdata, u_int line)
 	key = key_string_lookup_string(expanded);
 	free(expanded);
 	format_free(ft);
-	return key;
+	return (key);
 }
 
 static struct screen *
@@ -343,10 +343,10 @@ window_buffer_init(struct window_mode_entry *wme, struct cmd_find_state *fs,
 		data->key_format = xstrdup(WINDOW_BUFFER_DEFAULT_KEY_FORMAT);
 	else
 		data->key_format = xstrdup(args_get(args, 'K'));
-	if (args == NULL || args->argc == 0)
+	if (args == NULL || args_count(args) == 0)
 		data->command = xstrdup(WINDOW_BUFFER_DEFAULT_COMMAND);
 	else
-		data->command = xstrdup(args->argv[0]);
+		data->command = xstrdup(args_string(args, 0));
 
 	data->data = mode_tree_start(wp, args, window_buffer_build,
 	    window_buffer_draw, window_buffer_search, window_buffer_menu, NULL,
